@@ -21,6 +21,24 @@
 #include "mi.h"
 #include "private.h"
 
+#define NUM_ENABLES    (256u)
+
+_Static_assert(sizeof(struct nvme_mi_aem_supported_list_header) == 5,
+	"size_of_nvme_mi_aem_supported_list_header_is_not_5_bytes");
+_Static_assert(sizeof(struct nvme_mi_aem_supported_item) == 3,
+	"sizeof_nvme_mi_aem_supported_item_is_not_3_bytes");
+_Static_assert(sizeof(struct nvme_mi_aem_enable_item) == 3,
+	"size_of_ae_enable_item_t_is_not_3_bytes");
+_Static_assert(sizeof(struct nvme_mi_aem_enable_list_header) == 5,
+	"size_of_nvme_mi_aem_enable_list_header_is_not_5_bytes");
+_Static_assert(sizeof(struct nvme_mi_aem_occ_data) == 9,
+	"size_of_nvme_mi_aem_occ_data_is_not_9_bytes");
+_Static_assert(sizeof(struct nvme_mi_aem_occ_list_hdr) == 7,
+	"size_of_nvme_mi_aem_occ_list_hdr_is_not_7_bytes");
+
+static int nvme_mi_get_async_message(nvme_mi_ep_t ep,
+	struct nvme_mi_aem_msg *aem_msg, size_t *aem_msg_len);
+
 static const int default_timeout = 1000; /* milliseconds; endpoints may
 					    override */
 
@@ -414,10 +432,86 @@ static int nvme_mi_verify_resp_mic(struct nvme_mi_resp *resp)
 	return resp->mic != ~crc;
 }
 
+__attribute__((weak)) void *nvme_mi_submit_entry(__u8 type, const struct nvme_mi_msg_hdr *hdr,
+						 size_t hdr_len, const void *data, size_t data_len)
+{
+	return NULL;
+}
+
+__attribute__((weak)) void nvme_mi_submit_exit(__u8 type, const struct nvme_mi_msg_hdr *hdr,
+					       size_t hdr_len, const void *data, size_t data_len,
+					       void *user_data) { }
+
+
+int nvme_mi_async_read(nvme_mi_ep_t ep, struct nvme_mi_resp *resp)
+{
+	if (nvme_mi_ep_has_quirk(ep, NVME_QUIRK_MIN_INTER_COMMAND_TIME))
+		nvme_mi_record_resp_time(ep);
+
+	int rc = ep->transport->aem_read(ep, resp);
+
+	if (rc && errno == EWOULDBLOCK) {
+		//Sometimes we might get owned tag data from the wrong endpoint.
+		//This isn't an error, but we shouldn't process it here
+		resp->data_len = 0;//No data to process
+		return 0;
+	} else if (rc) {
+		nvme_msg(ep->root, LOG_INFO, "transport failure\n");
+		return rc;
+	}
+
+	if (ep->transport->mic_enabled) {
+		rc = nvme_mi_verify_resp_mic(resp);
+		if (rc) {
+			nvme_msg(ep->root, LOG_WARNING, "crc mismatch\n");
+			errno = EBADMSG;
+			return -1;
+		}
+	}
+
+	//TODO: There's a bunch of overlap with the nvme_mi_submit.  Maybe we make common helpers
+
+	/* basic response checks */
+	if (resp->hdr_len < sizeof(struct nvme_mi_msg_hdr)) {
+		nvme_msg(ep->root, LOG_DEBUG,
+			 "Bad response header len: %zd\n", resp->hdr_len);
+		errno = EPROTO;
+		return -1;
+	}
+
+	if (resp->hdr->type != NVME_MI_MSGTYPE_NVME) {
+		nvme_msg(ep->root, LOG_DEBUG,
+			 "Invalid message type 0x%02x\n", resp->hdr->type);
+		errno = EPROTO;
+		return -1;
+	}
+
+	if (!(resp->hdr->nmp & ~(NVME_MI_ROR_REQ << 7))) {
+		nvme_msg(ep->root, LOG_DEBUG,
+			 "ROR value in response indicates a response\n");
+		errno = EIO;
+		return -1;
+	}
+
+	if (!(resp->hdr->nmp & (NVME_MI_MT_AE << 3))) {
+		nvme_msg(ep->root, LOG_DEBUG,
+			 "NMIMT does not indicate AEM\n");
+		resp->data_len = 0;//No data to process
+		return 0;
+	}
+
+	return 0;
+}
+
+
 int nvme_mi_submit(nvme_mi_ep_t ep, struct nvme_mi_req *req,
 		   struct nvme_mi_resp *resp)
 {
 	int rc;
+	void *user_data;
+
+	user_data = nvme_mi_submit_entry(req->hdr->type, req->hdr, req->hdr_len, req->data,
+					 req->data_len);
 
 	if (req->hdr_len < sizeof(struct nvme_mi_msg_hdr)) {
 		errno = EINVAL;
@@ -429,11 +523,6 @@ int nvme_mi_submit(nvme_mi_ep_t ep, struct nvme_mi_req *req,
 		return -1;
 	}
 
-	if (req->data_len & 0x3) {
-		errno = EINVAL;
-		return -1;
-	}
-
 	if (resp->hdr_len < sizeof(struct nvme_mi_msg_hdr)) {
 		errno = EINVAL;
 		return -1;
@@ -502,10 +591,26 @@ int nvme_mi_submit(nvme_mi_ep_t ep, struct nvme_mi_req *req,
 		return -1;
 	}
 
+	nvme_mi_submit_exit(resp->hdr->type, resp->hdr, resp->hdr_len, resp->data, resp->data_len,
+			    user_data);
+
 	return 0;
 }
 
-static void nvme_mi_admin_init_req(struct nvme_mi_req *req,
+int nvme_mi_set_csi(nvme_mi_ep_t ep, uint8_t csi)
+{
+	uint8_t csi_bit = (csi) ? 1 : 0;
+
+	if (nvme_mi_ep_has_quirk(ep, NVME_QUIRK_CSI_1_NOT_SUPPORTED) && csi_bit)
+		return -1;
+
+	ep->csi = csi_bit;
+
+	return 0;
+}
+
+static void nvme_mi_admin_init_req(nvme_mi_ep_t ep,
+				   struct nvme_mi_req *req,
 				   struct nvme_mi_admin_req_hdr *hdr,
 				   __u16 ctrl_id, __u8 opcode)
 {
@@ -514,7 +619,8 @@ static void nvme_mi_admin_init_req(struct nvme_mi_req *req,
 
 	hdr->hdr.type = NVME_MI_MSGTYPE_NVME;
 	hdr->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-		(NVME_MI_MT_ADMIN << 3); /* we always use command slot 0 */
+		(NVME_MI_MT_ADMIN << 3) |
+		(ep->csi & 1);
 	hdr->opcode = opcode;
 	hdr->ctrl_id = cpu_to_le16(ctrl_id);
 
@@ -530,7 +636,8 @@ static void nvme_mi_admin_init_resp(struct nvme_mi_resp *resp,
 	resp->hdr_len = sizeof(*hdr);
 }
 
-static void nvme_mi_control_init_req(struct nvme_mi_req *req,
+static void nvme_mi_control_init_req(nvme_mi_ep_t ep,
+				    struct nvme_mi_req *req,
 				    struct nvme_mi_control_req *control_req,
 				    __u8 opcode, __u16 cpsp)
 {
@@ -539,7 +646,8 @@ static void nvme_mi_control_init_req(struct nvme_mi_req *req,
 
 	control_req->hdr.type = NVME_MI_MSGTYPE_NVME;
 	control_req->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-		(NVME_MI_MT_CONTROL << 3); /* we always use command slot 0 */
+		(NVME_MI_MT_CONTROL << 3) |
+		(ep->csi & 1);
 	control_req->opcode = opcode;
 	control_req->cpsp = cpu_to_le16(cpsp);
 
@@ -626,6 +734,28 @@ static int nvme_mi_control_parse_status(struct nvme_mi_resp *resp, __u16 *cpsr)
 	return control_resp->status;
 }
 
+static int nvme_mi_get_async_message(nvme_mi_ep_t ep,
+							struct nvme_mi_aem_msg *aem_msg,
+							size_t *aem_msg_len)
+{
+	struct nvme_mi_resp resp;
+
+	memset(&resp, 0, sizeof(resp));
+	resp.hdr = &aem_msg->hdr;
+	resp.hdr_len = sizeof(struct nvme_mi_msg_hdr);
+	resp.data = &aem_msg->occ_list_hdr;
+	resp.data_len = *aem_msg_len;
+
+	int rc = nvme_mi_async_read(ep, &resp);
+
+	if (rc)
+		return rc;
+
+	*aem_msg_len = resp.data_len;
+	return 0;
+}
+
+
 int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 		       struct nvme_mi_admin_req_hdr *admin_req,
 		       size_t req_data_size,
@@ -676,7 +806,9 @@ int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 
 	admin_req->hdr.type = NVME_MI_MSGTYPE_NVME;
 	admin_req->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-				(NVME_MI_MT_ADMIN << 3);
+			     (NVME_MI_MT_ADMIN << 3) |
+			     (ctrl->ep->csi & 1);
+
 	admin_req->ctrl_id = cpu_to_le16(ctrl->id);
 	memset(&req, 0, sizeof(req));
 	req.hdr = &admin_req->hdr;
@@ -684,8 +816,6 @@ int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 	req.data = admin_req + 1;
 	req.data_len = req_data_size;
 
-	nvme_mi_calc_req_mic(&req);
-
 	memset(&resp, 0, sizeof(resp));
 	resp.hdr = &admin_resp->hdr;
 	resp.hdr_len = sizeof(*admin_resp);
@@ -754,7 +884,7 @@ int nvme_mi_admin_admin_passthru(nvme_mi_ctrl_t ctrl, __u8 opcode, __u8 flags,
 			has_read_data = true;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id, opcode);
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id, opcode);
 	req_hdr.cdw1 = cpu_to_le32(nsid);
 	req_hdr.cdw2 = cpu_to_le32(cdw2);
 	req_hdr.cdw3 = cpu_to_le32(cdw3);
@@ -776,8 +906,6 @@ int nvme_mi_admin_admin_passthru(nvme_mi_ctrl_t ctrl, __u8 opcode, __u8 flags,
 		req.data_len = data_len;
 	}
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	if (has_read_data) {
@@ -831,7 +959,7 @@ int nvme_mi_admin_identify_partial(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id, nvme_admin_identify);
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id, nvme_admin_identify);
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
 	req_hdr.cdw10 = cpu_to_le32(args->cntid << 16 | args->cns);
 	req_hdr.cdw11 = cpu_to_le32((args->csi & 0xff) << 24 |
@@ -844,8 +972,6 @@ int nvme_mi_admin_identify_partial(nvme_mi_ctrl_t ctrl,
 		req_hdr.doff = cpu_to_le32(offset);
 	}
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 	resp.data = args->data;
 	resp.data_len = size;
@@ -877,7 +1003,7 @@ int nvme_mi_control(nvme_mi_ep_t ep, __u8 opcode,
 	struct nvme_mi_req req;
 	int rc = 0;
 
-	nvme_mi_control_init_req(&req, &control_req, opcode, cpsp);
+	nvme_mi_control_init_req(ep, &req, &control_req, opcode, cpsp);
 	nvme_mi_control_init_resp(&resp, &control_resp);
 
 	rc = nvme_mi_submit(ep, &req, &resp);
@@ -928,7 +1054,8 @@ static int __nvme_mi_admin_get_log(nvme_mi_ctrl_t ctrl,
 
 	ndw = (len >> 2) - 1;
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id, nvme_admin_get_log_page);
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
+		nvme_admin_get_log_page);
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
 	req_hdr.cdw10 = cpu_to_le32((ndw & 0xffff) << 16 |
 				    ((!final || args->rae) ? 1 : 0) << 15 |
@@ -944,8 +1071,6 @@ static int __nvme_mi_admin_get_log(nvme_mi_ctrl_t ctrl,
 	req_hdr.flags = 0x1;
 	req_hdr.dlen = cpu_to_le32(len & 0xffffffff);
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 	resp.data = args->log + offset;
 	resp.data_len = len;
@@ -1160,7 +1285,7 @@ int nvme_mi_admin_security_send(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_security_send);
 
 	req_hdr.cdw10 = cpu_to_le32(args->secp << 24 |
@@ -1175,8 +1300,6 @@ int nvme_mi_admin_security_send(nvme_mi_ctrl_t ctrl,
 	req.data = args->data;
 	req.data_len = args->data_len;
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1206,7 +1329,7 @@ int nvme_mi_admin_security_recv(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_security_recv);
 
 	req_hdr.cdw10 = cpu_to_le32(args->secp << 24 |
@@ -1219,8 +1342,6 @@ int nvme_mi_admin_security_recv(nvme_mi_ctrl_t ctrl,
 	req_hdr.flags = 0x1;
 	req_hdr.dlen = cpu_to_le32(args->data_len & 0xffffffff);
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 	resp.data = args->data;
 	resp.data_len = args->data_len;
@@ -1252,7 +1373,7 @@ int nvme_mi_admin_get_features(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_get_features);
 
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
@@ -1260,8 +1381,6 @@ int nvme_mi_admin_get_features(nvme_mi_ctrl_t ctrl,
 	req_hdr.cdw14 = cpu_to_le32(args->uuidx & 0x7f);
 	req_hdr.cdw11 = cpu_to_le32(args->cdw11);
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 	resp.data = args->data;
 	resp.data_len = args->data_len;
@@ -1279,6 +1398,37 @@ int nvme_mi_admin_get_features(nvme_mi_ctrl_t ctrl,
 	return 0;
 }
 
+static int __nvme_mi_admin_get_features(nvme_mi_ctrl_t ctrl, enum nvme_features_id fid,
+					enum nvme_get_features_sel sel, __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fid = fid,
+		.nsid = NVME_NSID_NONE,
+		.sel = sel,
+		.cdw11 = 0,
+		.uuidx = NVME_UUID_NONE,
+		.data_len = 0,
+		.data = NULL,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_mi_admin_get_features(ctrl, &args);
+}
+
+int nvme_mi_admin_get_features_arbitration(nvme_mi_ctrl_t ctrl, enum nvme_get_features_sel sel,
+					   __u32 *result)
+{
+	return __nvme_mi_admin_get_features(ctrl, NVME_FEAT_FID_ARBITRATION, sel, result);
+}
+
+int nvme_mi_admin_get_features_power_mgmt(nvme_mi_ctrl_t ctrl, enum nvme_get_features_sel sel,
+					  __u32 *result)
+{
+	return __nvme_mi_admin_get_features(ctrl, NVME_FEAT_FID_POWER_MGMT, sel, result);
+}
+
 int nvme_mi_admin_set_features(nvme_mi_ctrl_t ctrl,
 			       struct nvme_set_features_args *args)
 {
@@ -1293,7 +1443,7 @@ int nvme_mi_admin_set_features(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_set_features);
 
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
@@ -1308,8 +1458,6 @@ int nvme_mi_admin_set_features(nvme_mi_ctrl_t ctrl,
 	req.data_len = args->data_len;
 	req.data = args->data;
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1325,6 +1473,33 @@ int nvme_mi_admin_set_features(nvme_mi_ctrl_t ctrl,
 	return 0;
 }
 
+static int __nvme_mi_admin_set_features(nvme_mi_ctrl_t ctrl, __u8 fid, __u32 cdw11, bool save,
+					__u32 *result)
+{
+	struct nvme_set_features_args args = {
+		.args_size = sizeof(args),
+		.nsid = NVME_NSID_NONE,
+		.cdw11 = cdw11,
+		.cdw12 = 0,
+		.save = save,
+		.uuidx = NVME_UUID_NONE,
+		.cdw15 = 0,
+		.data_len = 0,
+		.data = NULL,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+	return nvme_mi_admin_set_features(ctrl, &args);
+}
+
+int nvme_mi_admin_set_features_power_mgmt(nvme_mi_ctrl_t ctrl, __u8 ps, __u8 wh, bool save,
+					  __u32 *result)
+{
+	__u32 value = NVME_SET(ps, FEAT_PWRMGMT_PS) | NVME_SET(wh, FEAT_PWRMGMT_WH);
+
+	return __nvme_mi_admin_set_features(ctrl, NVME_FEAT_FID_POWER_MGMT, value, save, result);
+}
+
 int nvme_mi_admin_ns_mgmt(nvme_mi_ctrl_t ctrl,
 			  struct nvme_ns_mgmt_args *args)
 {
@@ -1342,7 +1517,7 @@ int nvme_mi_admin_ns_mgmt(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_ns_mgmt);
 
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
@@ -1368,8 +1543,6 @@ int nvme_mi_admin_ns_mgmt(nvme_mi_ctrl_t ctrl,
 		req_hdr.flags = 0x1;
 	}
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1393,7 +1566,7 @@ int nvme_mi_admin_ns_attach(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_ns_attach);
 
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
@@ -1403,8 +1576,6 @@ int nvme_mi_admin_ns_attach(nvme_mi_ctrl_t ctrl,
 	req_hdr.dlen = cpu_to_le32(sizeof(*args->ctrlist));
 	req_hdr.flags = 0x1;
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1438,7 +1609,7 @@ int nvme_mi_admin_fw_download(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_fw_download);
 
 	req_hdr.cdw10 = cpu_to_le32((args->data_len >> 2) - 1);
@@ -1448,8 +1619,6 @@ int nvme_mi_admin_fw_download(nvme_mi_ctrl_t ctrl,
 	req_hdr.dlen = cpu_to_le32(args->data_len);
 	req_hdr.flags = 0x1;
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1473,15 +1642,13 @@ int nvme_mi_admin_fw_commit(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_fw_commit);
 
 	req_hdr.cdw10 = cpu_to_le32(((__u32)(args->bpid & 0x1) << 31) |
 				    ((args->action & 0x7) << 3) |
 				    ((args->slot & 0x7) << 0));
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1505,7 +1672,7 @@ int nvme_mi_admin_format_nvm(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_format_nvm);
 
 	req_hdr.cdw1 = cpu_to_le32(args->nsid);
@@ -1516,8 +1683,6 @@ int nvme_mi_admin_format_nvm(nvme_mi_ctrl_t ctrl,
 				    | ((args->mset & 0x1) << 4)
 				    | ((args->lbaf & 0xf) << 0));
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1541,7 +1706,7 @@ int nvme_mi_admin_sanitize_nvm(nvme_mi_ctrl_t ctrl,
 		return -1;
 	}
 
-	nvme_mi_admin_init_req(&req, &req_hdr, ctrl->id,
+	nvme_mi_admin_init_req(ctrl->ep, &req, &req_hdr, ctrl->id,
 			       nvme_admin_sanitize_nvm);
 
 	req_hdr.cdw10 = cpu_to_le32(((args->nodas ? 1 : 0) << 9)
@@ -1551,8 +1716,6 @@ int nvme_mi_admin_sanitize_nvm(nvme_mi_ctrl_t ctrl,
 				    | ((args->sanact & 0x7) << 0));
 	req_hdr.cdw11 = cpu_to_le32(args->ovrpat);
 
-	nvme_mi_calc_req_mic(&req);
-
 	nvme_mi_admin_init_resp(&resp, &resp_hdr);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
@@ -1562,6 +1725,25 @@ int nvme_mi_admin_sanitize_nvm(nvme_mi_ctrl_t ctrl,
 	return nvme_mi_admin_parse_status(&resp, args->result);
 }
 
+static void nvme_mi_mi_init_req(nvme_mi_ep_t ep,
+	struct nvme_mi_req *req,
+	struct nvme_mi_mi_req_hdr *hdr,
+	__u32 cdw0, __u8 opcode)
+{
+	memset(req, 0, sizeof(*req));
+	memset(hdr, 0, sizeof(*hdr));
+
+	hdr->hdr.type = NVME_MI_MSGTYPE_NVME;
+	hdr->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
+		(NVME_MI_MT_MI << 3) |
+		(ep->csi & 1);
+	hdr->opcode = opcode;
+	hdr->cdw0 = cpu_to_le32(cdw0);
+
+	req->hdr = &hdr->hdr;
+	req->hdr_len = sizeof(*hdr);
+}
+
 static int nvme_mi_read_data(nvme_mi_ep_t ep, __u32 cdw0,
 			     void *data, size_t *data_len)
 {
@@ -1571,16 +1753,8 @@ static int nvme_mi_read_data(nvme_mi_ep_t ep, __u32 cdw0,
 	struct nvme_mi_req req;
 	int rc;
 
-	memset(&req_hdr, 0, sizeof(req_hdr));
-	req_hdr.hdr.type = NVME_MI_MSGTYPE_NVME;
-	req_hdr.hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-		(NVME_MI_MT_MI << 3); /* we always use command slot 0 */
-	req_hdr.opcode = nvme_mi_mi_opcode_mi_data_read;
-	req_hdr.cdw0 = cpu_to_le32(cdw0);
-
-	memset(&req, 0, sizeof(req));
-	req.hdr = &req_hdr.hdr;
-	req.hdr_len = sizeof(req_hdr);
+	nvme_mi_mi_init_req(ep, &req, &req_hdr, cdw0,
+		nvme_mi_mi_opcode_mi_data_read);
 
 	memset(&resp, 0, sizeof(resp));
 	resp.hdr = &resp_hdr.hdr;
@@ -1626,22 +1800,10 @@ int nvme_mi_mi_xfer(nvme_mi_ep_t ep,
 		return -1;
 	}
 
-	/* request and response lengths & offset must be aligned */
-	if ((req_data_size & 0x3) ||
-	    (*resp_data_size & 0x3)) {
-		errno = EINVAL;
-		return -1;
-	}
-
-	/* bidirectional not permitted */
-	if (req_data_size && *resp_data_size) {
-		errno = EINVAL;
-		return -1;
-	}
-
 	mi_req->hdr.type = NVME_MI_MSGTYPE_NVME;
 	mi_req->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-				(NVME_MI_MT_MI << 3);
+			  (NVME_MI_MT_MI << 3) |
+			  (ep->csi & 1);
 
 	memset(&req, 0, sizeof(req));
 	req.hdr = &mi_req->hdr;
@@ -1649,8 +1811,6 @@ int nvme_mi_mi_xfer(nvme_mi_ep_t ep,
 	req.data = mi_req + 1;
 	req.data_len = req_data_size;
 
-	nvme_mi_calc_req_mic(&req);
-
 	memset(&resp, 0, sizeof(resp));
 	resp.hdr = &mi_resp->hdr;
 	resp.hdr_len = sizeof(*mi_resp);
@@ -1762,17 +1922,10 @@ int nvme_mi_mi_subsystem_health_status_poll(nvme_mi_ep_t ep, bool clear,
 	struct nvme_mi_req req;
 	int rc;
 
-	memset(&req_hdr, 0, sizeof(req_hdr));
-	req_hdr.hdr.type = NVME_MI_MSGTYPE_NVME;;
-	req_hdr.hdr.nmp = (NVME_MI_ROR_REQ << 7) |
-		(NVME_MI_MT_MI << 3);
-	req_hdr.opcode = nvme_mi_mi_opcode_subsys_health_status_poll;
+	nvme_mi_mi_init_req(ep, &req, &req_hdr, 0,
+		nvme_mi_mi_opcode_subsys_health_status_poll);
 	req_hdr.cdw1 = (clear ? 1 : 0) << 31;
 
-	memset(&req, 0, sizeof(req));
-	req.hdr = &req_hdr.hdr;
-	req.hdr_len = sizeof(req_hdr);
-
 	memset(&resp, 0, sizeof(resp));
 	resp.hdr = &resp_hdr.hdr;
 	resp.hdr_len = sizeof(resp_hdr);
@@ -1798,8 +1951,9 @@ int nvme_mi_mi_subsystem_health_status_poll(nvme_mi_ep_t ep, bool clear,
 	return 0;
 }
 
-int nvme_mi_mi_config_get(nvme_mi_ep_t ep, __u32 dw0, __u32 dw1,
-			  __u32 *nmresp)
+int nvme_mi_mi_config_set_get_ex(nvme_mi_ep_t ep, __u8 opcode, __u32 dw0,
+				__u32 dw1, void *data_out, size_t data_out_len,
+				void *data_in, size_t *data_in_len, __u32 *nmresp)
 {
 	struct nvme_mi_mi_resp_hdr resp_hdr;
 	struct nvme_mi_mi_req_hdr req_hdr;
@@ -1807,20 +1961,16 @@ int nvme_mi_mi_config_get(nvme_mi_ep_t ep, __u32 dw0, __u32 dw1,
 	struct nvme_mi_req req;
 	int rc;
 
-	memset(&req_hdr, 0, sizeof(req_hdr));
-	req_hdr.hdr.type = NVME_MI_MSGTYPE_NVME;
-	req_hdr.hdr.nmp = (NVME_MI_ROR_REQ << 7) | (NVME_MI_MT_MI << 3);
-	req_hdr.opcode = nvme_mi_mi_opcode_configuration_get;
-	req_hdr.cdw0 = cpu_to_le32(dw0);
+	nvme_mi_mi_init_req(ep, &req, &req_hdr, dw0, opcode);
 	req_hdr.cdw1 = cpu_to_le32(dw1);
-
-	memset(&req, 0, sizeof(req));
-	req.hdr = &req_hdr.hdr;
-	req.hdr_len = sizeof(req_hdr);
+	req.data = data_out;
+	req.data_len = data_out_len;
 
 	memset(&resp, 0, sizeof(resp));
 	resp.hdr = &resp_hdr.hdr;
 	resp.hdr_len = sizeof(resp_hdr);
+	resp.data = data_in;
+	resp.data_len = *data_in_len;
 
 	rc = nvme_mi_submit(ep, &req, &resp);
 	if (rc)
@@ -1829,46 +1979,122 @@ int nvme_mi_mi_config_get(nvme_mi_ep_t ep, __u32 dw0, __u32 dw1,
 	if (resp_hdr.status)
 		return resp_hdr.status;
 
-	*nmresp = resp_hdr.nmresp[0] |
-		  resp_hdr.nmresp[1] << 8 |
-		  resp_hdr.nmresp[2] << 16;
+	*data_in_len = resp.data_len;
+
+	if (nmresp) {
+		*nmresp = resp_hdr.nmresp[0] |
+		resp_hdr.nmresp[1] << 8 |
+		resp_hdr.nmresp[2] << 16;
+	}
 
 	return 0;
 }
 
+int nvme_mi_mi_config_get(nvme_mi_ep_t ep, __u32 dw0, __u32 dw1,
+			  __u32 *nmresp)
+{
+	size_t data_in_len = 0;
+
+	return nvme_mi_mi_config_set_get_ex(ep,
+					nvme_mi_mi_opcode_configuration_get,
+					dw0,
+					dw1,
+					NULL,
+					0,
+					NULL,
+					&data_in_len,
+					nmresp);
+}
+
 int nvme_mi_mi_config_set(nvme_mi_ep_t ep, __u32 dw0, __u32 dw1)
 {
-	struct nvme_mi_mi_resp_hdr resp_hdr;
-	struct nvme_mi_mi_req_hdr req_hdr;
-	struct nvme_mi_resp resp;
-	struct nvme_mi_req req;
-	int rc;
+	size_t data_in_len = 0;
 
-	memset(&req_hdr, 0, sizeof(req_hdr));
-	req_hdr.hdr.type = NVME_MI_MSGTYPE_NVME;
-	req_hdr.hdr.nmp = (NVME_MI_ROR_REQ << 7) | (NVME_MI_MT_MI << 3);
-	req_hdr.opcode = nvme_mi_mi_opcode_configuration_set;
-	req_hdr.cdw0 = cpu_to_le32(dw0);
-	req_hdr.cdw1 = cpu_to_le32(dw1);
+	return nvme_mi_mi_config_set_get_ex(ep,
+					nvme_mi_mi_opcode_configuration_set,
+					dw0,
+					dw1,
+					NULL,
+					0,
+					NULL,
+					&data_in_len,
+					NULL);
+}
 
-	memset(&req, 0, sizeof(req));
-	req.hdr = &req_hdr.hdr;
-	req.hdr_len = sizeof(req_hdr);
+int nvme_mi_mi_config_get_async_event(nvme_mi_ep_t ep,
+				__u8 *aeelver,
+				struct nvme_mi_aem_supported_list *list,
+				size_t *list_num_bytes)
+{
 
-	memset(&resp, 0, sizeof(resp));
-	resp.hdr = &resp_hdr.hdr;
-	resp.hdr_len = sizeof(resp_hdr);
+	__u32 dw0 = NVME_MI_CONFIG_AE;
+	__u32 aeelvertemp = 0;
+
+	int rc = nvme_mi_mi_config_set_get_ex(ep,
+					nvme_mi_mi_opcode_configuration_get,
+					dw0,
+					0,
+					NULL,
+					0,
+					list,
+					list_num_bytes,
+					&aeelvertemp);
 
-	rc = nvme_mi_submit(ep, &req, &resp);
 	if (rc)
 		return rc;
 
-	if (resp_hdr.status)
-		return resp_hdr.status;
+	*aeelver = 0x000F & aeelvertemp;
 
 	return 0;
 }
 
+int nvme_mi_mi_config_set_async_event(nvme_mi_ep_t ep,
+				bool envfa,
+				bool empfa,
+				bool encfa,
+				__u8 aemd,
+				__u8 aerd,
+				struct nvme_mi_aem_enable_list *enable_list,
+				size_t enable_list_size,
+				struct nvme_mi_aem_occ_list_hdr *occ_list,
+				size_t *occ_list_size)
+{
+
+	__u32 dw0 = ((__u32)envfa << 26) |
+				((__u32)empfa << 25) |
+				((__u32)encfa << 24) |
+				((__u32)aemd << 16)  |
+				((__u16) aerd << 8)  | NVME_MI_CONFIG_AE;
+
+	//Basic checks here on lengths
+	if (enable_list_size < sizeof(struct nvme_mi_aem_enable_list) ||
+		(sizeof(struct nvme_mi_aem_enable_list) +
+		 enable_list->hdr.numaee * sizeof(struct nvme_mi_aem_enable_item)
+		 > enable_list_size)
+	  ) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	//Some very baseic header checks
+	if (enable_list->hdr.aeelhl != sizeof(struct nvme_mi_aem_enable_list_header) ||
+		enable_list->hdr.aeelver != 0) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	return nvme_mi_mi_config_set_get_ex(ep,
+		nvme_mi_mi_opcode_configuration_set,
+		dw0,
+		0,
+		enable_list,
+		enable_list_size,
+		occ_list,
+		occ_list_size,
+		NULL);
+}
+
+
 void nvme_mi_close(nvme_mi_ep_t ep)
 {
 	struct nvme_mi_ctrl *ctrl, *tmp;
@@ -1974,3 +2200,516 @@ const char *nvme_mi_status_to_string(int status)
 
         return s;
 }
+
+bool nvme_mi_aem_aeei_get_aee(__le16 aeei)
+{
+	return !!(le16_to_cpu(aeei) & 0x8000);
+}
+
+__u8 nvme_mi_aem_aeei_get_aeeid(__le16 aeei)
+{
+	return (le16_to_cpu(aeei) & 0xFF);
+}
+
+void nvme_mi_aem_aeei_set_aeeid(struct nvme_mi_aem_enable_item *item, __u8 aeeid)
+{
+	__u16 temp = le16_to_cpu(item->aeei);
+
+	item->aeei = cpu_to_le16((temp & 0xFF00) | aeeid);
+}
+
+void nvme_mi_aem_aeei_set_aee(struct nvme_mi_aem_enable_item *item, bool enabled)
+{
+	__u16 temp = le16_to_cpu(item->aeei);
+	__u8 bit = (enabled) ? 1 : 0;
+
+	item->aeei = cpu_to_le16((temp & 0xFF) | (bit << 15));
+}
+
+bool nvme_mi_aem_aesi_get_aese(__le16 aesi)
+{
+	return !!(le16_to_cpu(aesi) & 0x8000);
+}
+
+__u8 nvme_mi_aem_aesi_get_aesid(__le16 aesi)
+{
+	return le16_to_cpu(aesi) & 0xff;
+}
+
+void nvme_mi_aem_aesi_set_aesid(struct nvme_mi_aem_supported_item *item, __u8 aesid)
+{
+	__u16 temp = le16_to_cpu(item->aesi);
+
+	item->aesi = cpu_to_le16((temp & 0xFF00) | aesid);
+}
+
+void nvme_mi_aem_aesi_set_aee(struct nvme_mi_aem_supported_item *item, bool enabled)
+{
+	__u16 temp = le16_to_cpu(item->aesi);
+	__u8 bit = (enabled) ? 1 : 0;
+
+	item->aesi = cpu_to_le16((temp & 0xFF) | (bit << 15));
+}
+
+__u8 nvme_mi_aem_aemti_get_aemgn(__u8 aemti)
+{
+	return aemti >> 3 & 0x1f;
+}
+
+__u32 nvme_mi_aem_aeolli_get_aeoltl(__u8 *aeolli)
+{
+	//First 23-bits contain the aeoltl
+	__u32 aeoltl = aeolli[0] | (aeolli[1] << 8) | (aeolli[2] << 16);
+
+	return aeoltl & 0x7FFFFF;
+}
+
+void nvme_mi_aem_aeolli_set_aeoltl(struct nvme_mi_aem_occ_list_hdr *hdr, __u32 aeoltl)
+{
+	hdr->aeolli[0] = aeoltl & 0xFF;
+	hdr->aeolli[1] = (aeoltl >> 8) & 0xFF;
+	hdr->aeolli[2] = (hdr->aeolli[2] & 0b10000000) | ((aeoltl >> 16) & 0x7F);
+}
+
+static int validate_enabled_list(struct nvme_mi_aem_supported_list *list, size_t len)
+{
+	if (list->hdr.aeslver != 0) {
+		errno = EPROTO;
+		return -1;
+	}
+	if (list->hdr.aeslhl != sizeof(struct nvme_mi_aem_supported_list)) {
+		errno = EPROTO;
+		return -1;
+	}
+	if (list->hdr.aest > len ||
+		list->hdr.aest !=
+		list->hdr.aeslhl + list->hdr.numaes * sizeof(struct nvme_mi_aem_supported_item)) {
+		errno = EPROTO;
+		return -1;
+	}
+	return 0;
+}
+static int validate_occ_list_update_ctx(
+							struct nvme_mi_aem_occ_list_hdr *occ_header,
+							size_t len,
+							struct nvme_mi_aem_ctx *ctx,
+							bool check_generation_num)
+{
+	//Make sure header fields have valid data
+	if (len < sizeof(*occ_header)) {
+		errno = EPROTO;
+		goto err_cleanup;
+	} else if (occ_header->aelver != 0 ||
+			   occ_header->aeolhl != sizeof(*occ_header)) {
+		//Make sure header is the right version and length
+		errno = EPROTO;
+		goto err_cleanup;
+	} else if (nvme_mi_aem_aeolli_get_aeoltl(occ_header->aeolli) > len) {
+		//Full length is bigger than the data that was received
+		errno = EPROTO;
+		goto err_cleanup;
+	} else if (check_generation_num &&
+		ctx->last_generation_num ==
+			(int) nvme_mi_aem_aemti_get_aemgn(occ_header->aemti)) {
+		//This is a duplicate and shouldn't be parsed.
+		//Let's just act like there's no updates
+		occ_header->numaeo = 0;
+	} else if (check_generation_num) {
+		ctx->last_generation_num =
+			nvme_mi_aem_aemti_get_aemgn(occ_header->aemti);
+	}
+
+	//Header is fine.  Let's go through the data
+	//First, we should update our context appropriately
+	ctx->occ_header = occ_header;
+
+	//Data starts after header
+	ctx->list_current = (struct nvme_mi_aem_occ_data *) (occ_header + 1);
+	ctx->list_current_index = 0;
+	ctx->list_start = ctx->list_current;
+
+	struct nvme_mi_aem_occ_data *current = ctx->list_current;
+	size_t bytes_so_far = ctx->occ_header->aeolhl;
+
+	for (int i = 0; i < occ_header->numaeo; i++) {
+		//Validate this item
+		if (current->aelhlen != sizeof(*current)) {
+			errno = EPROTO;
+			goto err_cleanup;
+		} else if (!ctx->callbacks.enabled_map.enabled[current->aeoui.aeoi]) {
+			//This is unexpected as this AE shouldn't be enabled
+			errno = EPROTO;
+			goto err_cleanup;
+		}
+
+		//Okay, check data lengths, including this header and the specific data(s)
+		uint32_t offset = sizeof(*current) + current->aeosil + current->aeovsil;
+
+		bytes_so_far += offset;
+		if (bytes_so_far > nvme_mi_aem_aeolli_get_aeoltl(occ_header->aeolli)) {
+			errno = EPROTO;
+			goto err_cleanup;
+		}
+
+		current = (struct nvme_mi_aem_occ_data *)((uint8_t *)current + offset);
+	}
+
+	return 0;
+
+err_cleanup:
+	return -1;
+}
+
+int nvme_mi_aem_get_fd(nvme_mi_ep_t ep)
+{
+	if (!ep || !ep->aem_ctx || !ep->transport || !ep->transport->aem_fd)
+		return -1;
+
+	return ep->transport->aem_fd(ep);
+}
+
+static void reset_list_info(struct nvme_mi_aem_ctx *ctx)
+{
+	//Reset context information
+	ctx->list_current_index = -1;
+	ctx->list_start = NULL;
+	ctx->list_current = NULL;
+	ctx->occ_header = NULL;
+}
+
+static int aem_sync(nvme_mi_ep_t ep,
+	bool envfa,
+	bool empfa,
+	bool encfa,
+	__u8 aemd,
+	__u8 aerd,
+	struct nvme_mi_aem_enable_item *items,
+	__u8 num_items,
+	struct nvme_mi_aem_occ_list_hdr *resp,
+	size_t *resp_len
+)
+{
+	size_t msg_len =
+		sizeof(struct nvme_mi_aem_enable_list_header) +
+		num_items * sizeof(struct nvme_mi_aem_enable_item);
+
+	struct nvme_mi_aem_enable_list_header *request = malloc(msg_len);
+
+	if (!request)
+		return -1;
+
+	request->aeelhl = sizeof(struct nvme_mi_aem_enable_list_header);
+	request->numaee = num_items;
+	request->aeelver = 0;
+	request->aeetl = msg_len;
+
+	//Data follows header
+	struct nvme_mi_aem_enable_item *msg_items = (struct nvme_mi_aem_enable_item *)(request + 1);
+
+	//Let's be explicit about what's enabled and what's not
+	for (int i = 0; i < num_items; i++) {
+		msg_items[i] = items[i];
+		msg_items[i].aeel = sizeof(msg_items[i]);
+	}
+
+	//Send it
+	int rc = nvme_mi_mi_config_set_async_event(ep,
+					envfa,
+					empfa,
+					encfa,
+					aemd,
+					aerd,
+					(struct nvme_mi_aem_enable_list *)request,
+					msg_len,
+					resp,
+					resp_len);
+
+	free(request);
+	return rc;
+}
+
+static int aem_disable_enabled(nvme_mi_ep_t ep)
+{
+	struct nvme_mi_aem_enabled_map already_enabled = {false};
+	uint8_t response_buffer[4096] = {0};
+	size_t response_len = sizeof(response_buffer);
+	struct nvme_mi_aem_occ_list_hdr *response =
+		(struct nvme_mi_aem_occ_list_hdr *)response_buffer;
+
+	// First, let's figure out if anything is already enabled that we need to
+    // disable
+	int rc = nvme_mi_aem_get_enabled(ep, &already_enabled);
+
+	if (rc)
+		return rc;
+
+	int sync_data_count = 0;
+
+	//Add the enabled items to the list of things to disable
+	struct nvme_mi_aem_enable_item sync_data[NUM_ENABLES] = {0};
+
+	for (int i = 0; i < NUM_ENABLES; i++) {
+		if (already_enabled.enabled[i]) {
+			nvme_mi_aem_aeei_set_aeeid(&sync_data[sync_data_count], i);
+			nvme_mi_aem_aeei_set_aee(&sync_data[sync_data_count], false);
+			sync_data_count++;
+		}
+	}
+
+	rc = aem_sync(ep, false, false, false, 1, 0,
+		sync_data, sync_data_count, response, &response_len);
+
+	if (rc)
+		return rc;
+
+	//Now, allow a purge of the aem fd because we could have
+	//received some events during this process
+	rc = ep->transport->aem_purge(ep);
+
+	return rc;
+}
+
+int nvme_mi_aem_enable(nvme_mi_ep_t ep,
+	struct nvme_mi_aem_config *config,
+	void *userdata)
+{
+	if (!ep || !config || !config->aem_handler)
+		return -1;
+
+	int rc = nvme_mi_aem_open(ep);
+
+	if (rc < 0)
+		return rc;
+
+	int sync_data_count = 0;
+	struct nvme_mi_aem_enable_item sync_data[NUM_ENABLES] = {0};
+
+	uint8_t response_buffer[4096] = {0};
+	size_t response_len = sizeof(response_buffer);
+	struct nvme_mi_aem_occ_list_hdr *response =
+		(struct nvme_mi_aem_occ_list_hdr *)response_buffer;
+
+	//It's possible we're already enabled
+	if (!ep->aem_ctx)
+		ep->aem_ctx = malloc(sizeof(*ep->aem_ctx));
+	if (!(ep->aem_ctx))
+		return -1;
+
+	memset(ep->aem_ctx, 0, sizeof(*ep->aem_ctx));
+	ep->aem_ctx->last_generation_num = -1;//Invalid
+	reset_list_info((ep->aem_ctx));
+	ep->aem_ctx->callbacks = *config;
+
+	rc = aem_disable_enabled(ep);
+	if (rc)
+		goto cleanup_ctx;
+
+	//Now, let's do a fresh enable of what's asked
+	for (int i = 0; i < NUM_ENABLES; i++) {
+		if (config->enabled_map.enabled[i]) {
+			nvme_mi_aem_aeei_set_aeeid(&sync_data[sync_data_count], i);
+			nvme_mi_aem_aeei_set_aee(&sync_data[sync_data_count], true);
+			sync_data_count++;
+		}
+	}
+
+	rc = aem_sync(ep, config->envfa, config->empfa,
+		config->encfa, config->aemd, config->aerd,
+		sync_data, sync_data_count, response, &response_len);
+	if (rc)
+		goto cleanup_ctx;
+
+	//Parse the response and fire events
+	rc = validate_occ_list_update_ctx(response,
+					response_len,
+					ep->aem_ctx,
+					false /*generation # shouldn't matter*/);
+	if (rc)
+		goto cleanup_ctx;
+
+	if (response->numaeo) {
+		//Return value unused here
+		config->aem_handler(ep, response->numaeo, userdata);
+	}
+
+cleanup_ctx:
+	// Clear these because they won't point to valid memory anymore
+	reset_list_info(ep->aem_ctx);
+
+	if (rc) {
+		free(ep->aem_ctx);
+		ep->aem_ctx = NULL;
+	}
+	return rc;
+}
+
+int nvme_mi_aem_get_enabled(nvme_mi_ep_t ep,
+	struct nvme_mi_aem_enabled_map *enabled_map)
+{
+	if (!ep || !enabled_map)
+		return -1;
+
+	int rc = 0;
+
+	unsigned char aeelver;
+	size_t ae_list_bytes = NUM_ENABLES * sizeof(struct nvme_mi_aem_supported_list);
+	struct nvme_mi_aem_supported_list *enabled_list = malloc(ae_list_bytes);
+
+	if (!enabled_list)
+		return -1;
+
+	rc = nvme_mi_mi_config_get_async_event(
+		ep, &aeelver, enabled_list, &ae_list_bytes);
+	if (rc)
+		goto cleanup;
+
+	rc = validate_enabled_list(enabled_list, ae_list_bytes);
+	if (rc)
+		goto cleanup;
+
+	memset(enabled_map, 0, sizeof(*enabled_map));
+
+	struct nvme_mi_aem_enable_item *items =
+		(struct nvme_mi_aem_enable_item *)(enabled_list + 1);
+
+	for (int i = 0; i < enabled_list->hdr.numaes; i++) {
+		__u8 aeeid = nvme_mi_aem_aeei_get_aeeid(items[i].aeei);
+		bool enabled = nvme_mi_aem_aeei_get_aee(items[i].aeei);
+
+		enabled_map->enabled[aeeid] = enabled;
+	}
+
+cleanup:
+	free(enabled_list);
+	return rc;
+}
+
+int nvme_mi_aem_disable(nvme_mi_ep_t ep)
+{
+	if (!ep)
+		return -1;
+
+	int rc = aem_disable_enabled(ep);
+
+	free(ep->aem_ctx);
+	ep->aem_ctx = NULL;
+
+	return rc;
+}
+
+/*When inside a aem_handler, call with the aem_ctx and struct will be populated with next
+ *event information.  Will return NULL when end of parsing (or error) is occurred.
+ *spec_info and vend_spec_info must be copied to persist as they will not be valid after
+ *the aem_handler has returned.
+ */
+struct nvme_mi_event *nvme_mi_aem_get_next_event(nvme_mi_ep_t ep)
+{
+	if (!ep || !ep->aem_ctx ||
+		!ep->aem_ctx->list_current ||
+		ep->aem_ctx->list_current_index == -1 ||
+	    !ep->aem_ctx->occ_header) {
+		return NULL;
+	}
+
+	if (ep->aem_ctx->occ_header->numaeo <= ep->aem_ctx->list_current_index)
+		return NULL;
+
+	struct nvme_mi_aem_ctx *aem_ctx = ep->aem_ctx;
+	struct nvme_mi_aem_occ_data *current = aem_ctx->list_current;
+
+	aem_ctx->event.aeoi = current->aeoui.aeoi;
+	aem_ctx->event.aessi = current->aeoui.aessi;
+	aem_ctx->event.aeocidi = current->aeoui.aeocidi;
+	aem_ctx->event.spec_info_len = current->aeosil;
+	aem_ctx->event.vend_spec_info_len = current->aeovsil;
+	//Now the pointers
+	aem_ctx->event.spec_info = ((uint8_t *)current + current->aelhlen);
+	aem_ctx->event.vend_spec_info =
+		((uint8_t *)aem_ctx->event.spec_info + aem_ctx->event.spec_info_len);
+
+	//Let's grab the next item (if there is any).
+	aem_ctx->list_current_index++;
+	aem_ctx->list_current =
+		(struct nvme_mi_aem_occ_data *)
+		((uint8_t *)aem_ctx->event.vend_spec_info + aem_ctx->event.vend_spec_info_len);
+
+	return &aem_ctx->event;
+}
+
+/* POLLIN has indicated events.  This function reads and processes them.
+ * A callback will likely be invoked.
+ */
+int nvme_mi_aem_process(nvme_mi_ep_t ep, void *userdata)
+{
+	int rc = 0;
+	uint8_t response_buffer[4096];
+	struct nvme_mi_aem_msg *response = (struct nvme_mi_aem_msg *)response_buffer;
+	size_t response_len = sizeof(response_buffer) - sizeof(struct nvme_mi_aem_msg);
+
+	if (!ep || !ep->aem_ctx)
+		return -1;
+
+	memset(response_buffer, 0, sizeof(response_buffer));
+
+	//Reset context information
+	reset_list_info(ep->aem_ctx);
+
+	rc = nvme_mi_get_async_message(ep, response, &response_len);
+	if (rc)
+		goto cleanup;
+
+	if (!response_len) {
+		//If no error and response length zero, we've likely received an owned
+		//tag message from a different endpoint than this path is responsible
+		//for monitoring.
+		goto cleanup;
+	}
+
+	//Parse the response and fire events
+	rc = validate_occ_list_update_ctx(&response->occ_list_hdr,
+						response_len,
+						ep->aem_ctx,
+						true /*Ensure unique generation number*/);
+	if (rc)
+		goto cleanup;
+
+	if (response->occ_list_hdr.numaeo) {
+		enum nvme_mi_aem_handler_next_action action =
+			ep->aem_ctx->callbacks.aem_handler(ep,
+								response->occ_list_hdr.numaeo,
+								userdata);
+
+		reset_list_info(ep->aem_ctx);
+
+		if (action == NVME_MI_AEM_HNA_ACK) {
+			response_len = sizeof(response_buffer);
+
+			rc = nvme_mi_aem_ack(ep, &response->occ_list_hdr, &response_len);
+			if (rc)
+				goto cleanup;
+
+			//The Ack is not guaranteed to have data
+			if (response_len && response->occ_list_hdr.numaeo) {
+				rc = validate_occ_list_update_ctx(&response->occ_list_hdr,
+										response_len,
+										ep->aem_ctx,
+										true);
+				//Callbacks based on ack
+				if (rc == 0 && response->occ_list_hdr.numaeo) {
+					//Return value unused here
+					ep->aem_ctx->callbacks.aem_handler(ep,
+							response->occ_list_hdr.numaeo,
+							userdata);
+				}
+			}
+		}
+	} else {
+		//This is unexpected unless we have duplicates.  But those shouldn't be acked
+	}
+
+cleanup:
+	reset_list_info(ep->aem_ctx);
+	return rc;
+}
+
+
