@@ -66,11 +66,9 @@ static struct nvme_host *default_host;
 static void __nvme_free_host(nvme_host_t h);
 static void __nvme_free_ctrl(nvme_ctrl_t c);
 static int nvme_subsystem_scan_namespace(nvme_root_t r,
-		struct nvme_subsystem *s, char *name,
-		nvme_scan_filter_t f, void *f_args);
+					 struct nvme_subsystem *s, char *name);
 static int nvme_init_subsystem(nvme_subsystem_t s, const char *name);
-static int nvme_scan_subsystem(nvme_root_t r, const char *name,
-			       nvme_scan_filter_t f, void *f_args);
+static int nvme_scan_subsystem(nvme_root_t r, const char *name);
 static int nvme_ctrl_scan_namespace(nvme_root_t r, struct nvme_ctrl *c,
 				    char *name);
 static int nvme_ctrl_scan_path(nvme_root_t r, struct nvme_ctrl *c, char *name);
@@ -219,6 +217,68 @@ nvme_host_t nvme_default_host(nvme_root_t r)
 	return h;
 }
 
+static void nvme_filter_subsystem(nvme_root_t r, nvme_subsystem_t s,
+				  nvme_scan_filter_t f, void *f_args)
+{
+	if (f(s, NULL, NULL, f_args))
+		return;
+
+	nvme_msg(r, LOG_DEBUG, "filter out subsystem %s\n",
+		 nvme_subsystem_get_name(s));
+	nvme_free_subsystem(s);
+}
+
+static void nvme_filter_ns(nvme_root_t r, nvme_ns_t n,
+			   nvme_scan_filter_t f, void *f_args)
+{
+	if (f(NULL, NULL, n, f_args))
+		return;
+
+	nvme_msg(r, LOG_DEBUG, "filter out namespace %s\n",
+		 nvme_ns_get_name(n));
+	nvme_free_ns(n);
+}
+
+static void nvme_filter_ctrl(nvme_root_t r, nvme_ctrl_t c,
+			     nvme_scan_filter_t f, void *f_args)
+{
+	if (f(NULL, c, NULL, f_args))
+		return;
+
+	nvme_msg(r, LOG_DEBUG, "filter out controller %s\n",
+		 nvme_ctrl_get_name(c));
+	nvme_free_ctrl(c);
+}
+
+static void nvme_filter_tree(nvme_root_t r, nvme_scan_filter_t f, void *f_args)
+{
+	nvme_host_t h, _h;
+	nvme_subsystem_t s, _s;
+	nvme_ns_t n, _n;
+	nvme_path_t p, _p;
+	nvme_ctrl_t c, _c;
+
+	if (!f)
+		return;
+
+	nvme_for_each_host_safe(r, h, _h) {
+		nvme_for_each_subsystem_safe(h, s, _s) {
+			nvme_subsystem_for_each_ctrl_safe(s, c, _c)
+				nvme_filter_ctrl(r, c, f, f_args);
+
+			nvme_subsystem_for_each_ns_safe(s, n, _n) {
+				nvme_namespace_for_each_path_safe(n, p, _p) {
+					nvme_filter_ctrl(r, nvme_path_get_ctrl(p),
+							 f, f_args);
+				}
+				nvme_filter_ns(r, n, f, f_args);
+			}
+
+			nvme_filter_subsystem(r, s, f, f_args);
+		}
+	}
+}
+
 int nvme_scan_topology(struct nvme_root *r, nvme_scan_filter_t f, void *f_args)
 {
 	_cleanup_dirents_ struct dirents subsys = {}, ctrls = {};
@@ -241,11 +301,6 @@ int nvme_scan_topology(struct nvme_root *r, nvme_scan_filter_t f, void *f_args)
 				 ctrls.ents[i]->d_name, strerror(errno));
 			continue;
 		}
-		if ((f) && !f(NULL, c, NULL, f_args)) {
-			nvme_msg(r, LOG_DEBUG, "filter out controller %s\n",
-				 ctrls.ents[i]->d_name);
-			nvme_free_ctrl(c);
-		}
 	}
 
 	subsys.num = nvme_scan_subsystems(&subsys.ents);
@@ -256,8 +311,7 @@ int nvme_scan_topology(struct nvme_root *r, nvme_scan_filter_t f, void *f_args)
 	}
 
 	for (i = 0; i < subsys.num; i++) {
-		ret = nvme_scan_subsystem(
-			r, subsys.ents[i]->d_name, f, f_args);
+		ret = nvme_scan_subsystem(r, subsys.ents[i]->d_name);
 		if (ret < 0) {
 			nvme_msg(r, LOG_DEBUG,
 				 "failed to scan subsystem %s: %s\n",
@@ -265,6 +319,12 @@ int nvme_scan_topology(struct nvme_root *r, nvme_scan_filter_t f, void *f_args)
 		}
 	}
 
+	/*
+	 * Filter the tree after it has been fully populated and
+	 * updated
+	 */
+	nvme_filter_tree(r, f, f_args);
+
 	return 0;
 }
 
@@ -361,10 +421,9 @@ const char *nvme_root_get_application(nvme_root_t r)
 
 void nvme_root_set_application(nvme_root_t r, const char *a)
 {
-	if (r->application) {
-		free(r->application);
-		r->application = NULL;
-	}
+	free(r->application);
+	r->application = NULL;
+
 	if (a)
 		r->application = strdup(a);
 }
@@ -406,10 +465,9 @@ const char *nvme_host_get_hostsymname(nvme_host_t h)
 
 void nvme_host_set_hostsymname(nvme_host_t h, const char *hostsymname)
 {
-	if (h->hostsymname) {
-		free(h->hostsymname);
-		h->hostsymname = NULL;
-	}
+	free(h->hostsymname);
+	h->hostsymname = NULL;
+
 	if (hostsymname)
 		h->hostsymname = strdup(hostsymname);
 }
@@ -421,10 +479,9 @@ const char *nvme_host_get_dhchap_key(nvme_host_t h)
 
 void nvme_host_set_dhchap_key(nvme_host_t h, const char *key)
 {
-	if (h->dhchap_key) {
-		free(h->dhchap_key);
-		h->dhchap_key = NULL;
-	}
+	free(h->dhchap_key);
+	h->dhchap_key = NULL;
+
 	if (key)
 		h->dhchap_key = strdup(key);
 }
@@ -468,14 +525,11 @@ void nvme_free_tree(nvme_root_t r)
 	if (!r)
 		return;
 
-	if (r->options)
-		free(r->options);
+	free(r->options);
 	nvme_for_each_host_safe(r, h, _h)
 		__nvme_free_host(h);
-	if (r->config_file)
-		free(r->config_file);
-	if (r->application)
-		free(r->application);
+	free(r->config_file);
+	free(r->application);
 	free(r);
 }
 
@@ -514,10 +568,9 @@ const char *nvme_subsystem_get_application(nvme_subsystem_t s)
 
 void nvme_subsystem_set_application(nvme_subsystem_t s, const char *a)
 {
-	if (s->application) {
-		free(s->application);
-		s->application = NULL;
-	}
+	free(s->application);
+	s->application = NULL;
+
 	if (a)
 		s->application = strdup(a);
 }
@@ -532,6 +585,11 @@ const char *nvme_subsystem_get_model(nvme_subsystem_t s)
 	return s->model;
 }
 
+const char *nvme_subsystem_get_serial(nvme_subsystem_t s)
+{
+	return s->serial;
+}
+
 const char *nvme_subsystem_get_fw_rev(nvme_subsystem_t s)
 {
 	return s->firmware;
@@ -564,12 +622,12 @@ nvme_ns_t nvme_subsystem_next_ns(nvme_subsystem_t s, nvme_ns_t n)
 
 nvme_path_t nvme_namespace_first_path(nvme_ns_t ns)
 {
-	return list_top(&ns->paths, struct nvme_path, nentry);
+	return list_top(&ns->head->paths, struct nvme_path, nentry);
 }
 
 nvme_path_t nvme_namespace_next_path(nvme_ns_t ns, nvme_path_t p)
 {
-	return p ? list_next(&ns->paths, p, nentry) : NULL;
+	return p ? list_next(&ns->head->paths, p, nentry) : NULL;
 }
 
 static void __nvme_free_ns(struct nvme_ns *n)
@@ -579,6 +637,8 @@ static void __nvme_free_ns(struct nvme_ns *n)
 	free(n->generic_name);
 	free(n->name);
 	free(n->sysfs_dir);
+	free(n->head->sysfs_dir);
+	free(n->head);
 	free(n);
 }
 
@@ -600,22 +660,15 @@ static void __nvme_free_subsystem(struct nvme_subsystem *s)
 	nvme_subsystem_for_each_ns_safe(s, n, _n)
 		__nvme_free_ns(n);
 
-	if (s->name)
-		free(s->name);
+	free(s->name);
 	free(s->sysfs_dir);
 	free(s->subsysnqn);
-	if (s->model)
-		free(s->model);
-	if (s->serial)
-		free(s->serial);
-	if (s->firmware)
-		free(s->firmware);
-	if (s->subsystype)
-		free(s->subsystype);
-	if (s->application)
-		free(s->application);
-	if (s->iopolicy)
-		free(s->iopolicy);
+	free(s->model);
+	free(s->serial);
+	free(s->firmware);
+	free(s->subsystype);
+	free(s->application);
+	free(s->iopolicy);
 	free(s);
 }
 
@@ -692,10 +745,8 @@ static void __nvme_free_host(struct nvme_host *h)
 	nvme_for_each_subsystem_safe(h, s, _s)
 		__nvme_free_subsystem(s);
 	free(h->hostnqn);
-	if (h->hostid)
-		free(h->hostid);
-	if (h->dhchap_key)
-		free(h->dhchap_key);
+	free(h->hostid);
+	free(h->dhchap_key);
 	nvme_host_set_hostsymname(h, NULL);
 	h->r->modified = true;
 	free(h);
@@ -745,8 +796,7 @@ struct nvme_host *nvme_lookup_host(nvme_root_t r, const char *hostnqn,
 	return h;
 }
 
-static int nvme_subsystem_scan_namespaces(nvme_root_t r, nvme_subsystem_t s,
-		nvme_scan_filter_t f, void *f_args)
+static int nvme_subsystem_scan_namespaces(nvme_root_t r, nvme_subsystem_t s)
 {
 	_cleanup_dirents_ struct dirents namespaces = {};
 	int i, ret;
@@ -767,7 +817,7 @@ static int nvme_subsystem_scan_namespaces(nvme_root_t r, nvme_subsystem_t s,
 
 	for (i = 0; i < namespaces.num; i++) {
 		ret = nvme_subsystem_scan_namespace(r, s,
-				namespaces.ents[i]->d_name, f, f_args);
+				namespaces.ents[i]->d_name);
 		if (ret < 0)
 			nvme_msg(r, LOG_DEBUG,
 				 "failed to scan namespace %s: %s\n",
@@ -805,20 +855,7 @@ static int nvme_init_subsystem(nvme_subsystem_t s, const char *name)
 	return 0;
 }
 
-static bool __nvme_scan_subsystem(struct nvme_root *r, nvme_subsystem_t s,
-				  nvme_scan_filter_t f, void *f_args)
-{
-	if (f && !f(s, NULL, NULL, f_args)) {
-		nvme_msg(r, LOG_DEBUG, "filter out subsystem %s\n", s->name);
-		__nvme_free_subsystem(s);
-		return false;
-	}
-	nvme_subsystem_scan_namespaces(r, s, f, f_args);
-	return true;
-}
-
-static int nvme_scan_subsystem(struct nvme_root *r, const char *name,
-		nvme_scan_filter_t f, void *f_args)
+static int nvme_scan_subsystem(struct nvme_root *r, const char *name)
 {
 	struct nvme_subsystem *s = NULL, *_s;
 	_cleanup_free_ char *path = NULL, *subsysnqn = NULL;
@@ -846,7 +883,7 @@ static int nvme_scan_subsystem(struct nvme_root *r, const char *name,
 				continue;
 			if (strcmp(_s->name, name))
 				continue;
-			if (!__nvme_scan_subsystem(r, _s, f, f_args)) {
+			if (!nvme_subsystem_scan_namespaces(r, _s)) {
 				errno = EINVAL;
 				return -1;
 			}
@@ -867,7 +904,7 @@ static int nvme_scan_subsystem(struct nvme_root *r, const char *name,
 			errno = ENOMEM;
 			return -1;
 		}
-		if (!__nvme_scan_subsystem(r, s, f, f_args)) {
+		if (!nvme_subsystem_scan_namespaces(r, s)) {
 			errno = EINVAL;
 			return -1;
 		}
@@ -901,11 +938,28 @@ const char *nvme_path_get_name(nvme_path_t p)
 	return p->name;
 }
 
+int nvme_path_get_queue_depth(nvme_path_t p)
+{
+	_cleanup_free_ char *queue_depth = NULL;
+
+	queue_depth = nvme_get_path_attr(p, "queue_depth");
+	if (queue_depth) {
+		sscanf(queue_depth, "%d", &p->queue_depth);
+	}
+
+	return p->queue_depth;
+}
+
 const char *nvme_path_get_ana_state(nvme_path_t p)
 {
 	return p->ana_state;
 }
 
+const char *nvme_path_get_numa_nodes(nvme_path_t p)
+{
+	return p->numa_nodes;
+}
+
 void nvme_free_path(struct nvme_path *p)
 {
 	list_del_init(&p->entry);
@@ -913,32 +967,14 @@ void nvme_free_path(struct nvme_path *p)
 	free(p->name);
 	free(p->sysfs_dir);
 	free(p->ana_state);
+	free(p->numa_nodes);
 	free(p);
 }
 
-static void nvme_subsystem_set_path_ns(nvme_subsystem_t s, nvme_path_t p)
-{
-	char n_name[32] = { };
-	int i, c, nsid, ret;
-	nvme_ns_t n;
-
-	ret = sscanf(nvme_path_get_name(p), "nvme%dc%dn%d", &i, &c, &nsid);
-	if (ret != 3)
-		return;
-
-	sprintf(n_name, "nvme%dn%d", i, nsid);
-	nvme_subsystem_for_each_ns(s, n) {
-		if (!strcmp(n_name, nvme_ns_get_name(n))) {
-			list_add_tail(&n->paths, &p->nentry);
-			p->n = n;
-		}
-	}
-}
-
 static int nvme_ctrl_scan_path(nvme_root_t r, struct nvme_ctrl *c, char *name)
 {
 	struct nvme_path *p;
-	_cleanup_free_ char *path = NULL, *grpid = NULL;
+	_cleanup_free_ char *path = NULL, *grpid = NULL, *queue_depth = NULL;
 	int ret;
 
 	nvme_msg(r, LOG_DEBUG, "scan controller %s path %s\n",
@@ -967,13 +1003,21 @@ static int nvme_ctrl_scan_path(nvme_root_t r, struct nvme_ctrl *c, char *name)
 	if (!p->ana_state)
 		p->ana_state = strdup("optimized");
 
+	p->numa_nodes = nvme_get_path_attr(p, "numa_nodes");
+	if (!p->numa_nodes)
+		p->numa_nodes = strdup("-1");
+
 	grpid = nvme_get_path_attr(p, "ana_grpid");
 	if (grpid) {
 		sscanf(grpid, "%d", &p->grpid);
 	}
 
+	queue_depth = nvme_get_path_attr(p, "queue_depth");
+	if (queue_depth) {
+		sscanf(queue_depth, "%d", &p->queue_depth);
+	}
+
 	list_node_init(&p->nentry);
-	nvme_subsystem_set_path_ns(c->s, p);
 	list_node_init(&p->entry);
 	list_add_tail(&c->paths, &p->entry);
 	return 0;
@@ -1071,8 +1115,7 @@ const char *nvme_ctrl_get_state(nvme_ctrl_t c)
 	char *state = c->state;
 
 	c->state = nvme_get_ctrl_attr(c, "state");
-	if (state)
-		free(state);
+	free(state);
 	return c->state;
 }
 
@@ -1138,10 +1181,9 @@ const char *nvme_ctrl_get_cntlid(nvme_ctrl_t c)
 
 void nvme_ctrl_set_dhchap_host_key(nvme_ctrl_t c, const char *key)
 {
-	if (c->dhchap_key) {
-		free(c->dhchap_key);
-		c->dhchap_key = NULL;
-	}
+	free(c->dhchap_key);
+	c->dhchap_key = NULL;
+
 	if (key)
 		c->dhchap_key = strdup(key);
 }
@@ -1153,10 +1195,9 @@ const char *nvme_ctrl_get_dhchap_key(nvme_ctrl_t c)
 
 void nvme_ctrl_set_dhchap_key(nvme_ctrl_t c, const char *key)
 {
-	if (c->dhchap_ctrl_key) {
-		free(c->dhchap_ctrl_key);
-		c->dhchap_ctrl_key = NULL;
-	}
+	free(c->dhchap_ctrl_key);
+	c->dhchap_ctrl_key = NULL;
+
 	if (key)
 		c->dhchap_ctrl_key = strdup(key);
 }
@@ -1168,10 +1209,9 @@ const char *nvme_ctrl_get_keyring(nvme_ctrl_t c)
 
 void nvme_ctrl_set_keyring(nvme_ctrl_t c, const char *keyring)
 {
-	if (c->keyring) {
-		free(c->keyring);
-		c->keyring = NULL;
-	}
+	free(c->keyring);
+	c->keyring = NULL;
+
 	if (keyring)
 		c->keyring = strdup(keyring);
 }
@@ -1183,10 +1223,9 @@ const char *nvme_ctrl_get_tls_key_identity(nvme_ctrl_t c)
 
 void nvme_ctrl_set_tls_key_identity(nvme_ctrl_t c, const char *identity)
 {
-	if (c->tls_key_identity) {
-		free(c->tls_key_identity);
-		c->tls_key_identity = NULL;
-	}
+	free(c->tls_key_identity);
+	c->tls_key_identity = NULL;
+
 	if (identity)
 		c->tls_key_identity = strdup(identity);
 }
@@ -1198,10 +1237,9 @@ const char *nvme_ctrl_get_tls_key(nvme_ctrl_t c)
 
 void nvme_ctrl_set_tls_key(nvme_ctrl_t c, const char *key)
 {
-	if (c->tls_key) {
-		free(c->tls_key);
-		c->tls_key = NULL;
-	}
+	free(c->tls_key);
+	c->tls_key = NULL;
+
 	if (key)
 		c->tls_key = strdup(key);
 }
@@ -1272,7 +1310,7 @@ nvme_path_t nvme_ctrl_next_path(nvme_ctrl_t c, nvme_path_t p)
 }
 
 #define FREE_CTRL_ATTR(a) \
-	do { if (a) { free(a); (a) = NULL; } } while (0)
+	do { free(a); (a) = NULL; } while (0)
 void nvme_deconfigure_ctrl(nvme_ctrl_t c)
 {
 	nvme_ctrl_release_fd(c);
@@ -1709,8 +1747,8 @@ static ctrl_match_t _candidate_init(struct candidate_args *candidate,
 	candidate->trsvcid = trsvcid;
 	candidate->transport = transport;
 	candidate->subsysnqn = subsysnqn;
-	candidate->host_iface = host_iface;
-	candidate->host_traddr = host_traddr;
+	candidate->host_iface = streqcase0(host_iface, "none") ? NULL : host_iface;
+	candidate->host_traddr = streqcase0(host_traddr, "none") ? NULL : host_traddr;
 
 	if (streq0(subsysnqn, NVME_DISC_SUBSYS_NAME)) {
 		/* Since TP8013, the NQN of discovery controllers can be the
@@ -1946,9 +1984,7 @@ static void nvme_read_sysfs_dhchap(nvme_root_t r, nvme_ctrl_t c)
 	char *host_key, *ctrl_key;
 
 	host_key = nvme_get_ctrl_attr(c, "dhchap_secret");
-	if (host_key && c->s && c->s->h && c->s->h->dhchap_key &&
-			(!strcmp(c->s->h->dhchap_key, host_key) ||
-			 !strcmp("none", host_key))) {
+	if (host_key && !strcmp(host_key, "none")) {
 		free(host_key);
 		host_key = NULL;
 	}
@@ -2004,11 +2040,30 @@ static void nvme_read_sysfs_tls(nvme_root_t r, nvme_ctrl_t c)
 	free(key);
 }
 
-static int nvme_configure_ctrl(nvme_root_t r, nvme_ctrl_t c, const char *path,
-			       const char *name)
+static int nvme_reconfigure_ctrl(nvme_root_t r, nvme_ctrl_t c, const char *path,
+				 const char *name)
 {
 	DIR *d;
 
+	/*
+	 * It's necesssary to release any resources first because a ctrl
+	 * can be reused.
+	 */
+	nvme_ctrl_release_fd(c);
+	FREE_CTRL_ATTR(c->name);
+	FREE_CTRL_ATTR(c->sysfs_dir);
+	FREE_CTRL_ATTR(c->firmware);
+	FREE_CTRL_ATTR(c->model);
+	FREE_CTRL_ATTR(c->state);
+	FREE_CTRL_ATTR(c->numa_node);
+	FREE_CTRL_ATTR(c->queue_count);
+	FREE_CTRL_ATTR(c->serial);
+	FREE_CTRL_ATTR(c->sqsize);
+	FREE_CTRL_ATTR(c->cntrltype);
+	FREE_CTRL_ATTR(c->cntlid);
+	FREE_CTRL_ATTR(c->dctype);
+	FREE_CTRL_ATTR(c->phy_slot);
+
 	d = opendir(path);
 	if (!d) {
 		nvme_msg(r, LOG_ERR, "Failed to open ctrl dir %s, error %d\n",
@@ -2018,9 +2073,8 @@ static int nvme_configure_ctrl(nvme_root_t r, nvme_ctrl_t c, const char *path,
 	}
 	closedir(d);
 
-	c->fd = -1;
 	c->name = strdup(name);
-	c->sysfs_dir = (char *)path;
+	c->sysfs_dir = strdup(path);
 	c->firmware = nvme_get_ctrl_attr(c, "firmware_rev");
 	c->model = nvme_get_ctrl_attr(c, "model");
 	c->state = nvme_get_ctrl_attr(c, "state");
@@ -2041,10 +2095,8 @@ static int nvme_configure_ctrl(nvme_root_t r, nvme_ctrl_t c, const char *path,
 
 int nvme_init_ctrl(nvme_host_t h, nvme_ctrl_t c, int instance)
 {
-	_cleanup_free_ char *subsys_name = NULL;
-	_cleanup_free_ char *name = NULL;
+	_cleanup_free_ char *subsys_name = NULL, *name = NULL, *path = NULL;
 	nvme_subsystem_t s;
-	char *path;
 	int ret;
 
 	ret = asprintf(&name, "nvme%d", instance);
@@ -2058,11 +2110,9 @@ int nvme_init_ctrl(nvme_host_t h, nvme_ctrl_t c, int instance)
 		return ret;
 	}
 
-	ret = nvme_configure_ctrl(h->r, c, path, name);
-	if (ret < 0) {
-		free(path);
+	ret = nvme_reconfigure_ctrl(h->r, c, path, name);
+	if (ret < 0)
 		return ret;
-	}
 
 	c->address = nvme_get_attr(path, "address");
 	if (!c->address && strcmp(c->transport, "loop")) {
@@ -2093,12 +2143,11 @@ int nvme_init_ctrl(nvme_host_t h, nvme_ctrl_t c, int instance)
 static nvme_ctrl_t nvme_ctrl_alloc(nvme_root_t r, nvme_subsystem_t s,
 				   const char *path, const char *name)
 {
-	nvme_ctrl_t c, p;
-	_cleanup_free_ char *addr = NULL, *address = NULL;
-	char *a = NULL, *e = NULL;
-	_cleanup_free_ char *transport = NULL;
-	char *traddr = NULL, *trsvcid = NULL;
+	_cleanup_free_ char *addr = NULL, *address = NULL, *transport = NULL;
 	char *host_traddr = NULL, *host_iface = NULL;
+	char *traddr = NULL, *trsvcid = NULL;
+	char *a = NULL, *e = NULL;
+	nvme_ctrl_t c, p;
 	int ret;
 
 	transport = nvme_get_attr(path, "transport");
@@ -2180,22 +2229,22 @@ skip_address:
 		errno = ENODEV;
 		return NULL;
 	}
-	c->address = addr;
-	addr = NULL;
+	FREE_CTRL_ATTR(c->address);
+	c->address = strdup(addr);
 	if (s->subsystype && !strcmp(s->subsystype, "discovery"))
 		c->discovery_ctrl = true;
-	ret = nvme_configure_ctrl(r, c, path, name);
+	ret = nvme_reconfigure_ctrl(r, c, path, name);
 	return (ret < 0) ? NULL : c;
 }
 
 nvme_ctrl_t nvme_scan_ctrl(nvme_root_t r, const char *name)
 {
+	_cleanup_free_ char *subsysnqn = NULL, *subsysname = NULL;
+	_cleanup_free_ char *hostnqn = NULL, *hostid = NULL;
+	_cleanup_free_ char *path = NULL;
 	nvme_host_t h;
 	nvme_subsystem_t s;
 	nvme_ctrl_t c;
-	_cleanup_free_ char *path = NULL;
-	_cleanup_free_ char *hostnqn = NULL, *hostid = NULL;
-	_cleanup_free_ char *subsysnqn = NULL, *subsysname = NULL;
 	int ret;
 
 	nvme_msg(r, LOG_DEBUG, "scan controller %s\n", name);
@@ -2209,8 +2258,7 @@ nvme_ctrl_t nvme_scan_ctrl(nvme_root_t r, const char *name)
 	hostid = nvme_get_attr(path, "hostid");
 	h = nvme_lookup_host(r, hostnqn, hostid);
 	if (h) {
-		if (h->dhchap_key)
-			free(h->dhchap_key);
+		free(h->dhchap_key);
 		h->dhchap_key = nvme_get_attr(path, "dhchap_secret");
 		if (h->dhchap_key && !strcmp(h->dhchap_key, "none")) {
 			free(h->dhchap_key);
@@ -2249,9 +2297,8 @@ nvme_ctrl_t nvme_scan_ctrl(nvme_root_t r, const char *name)
 	if (!c)
 		return NULL;
 
-	path = NULL;
-	nvme_ctrl_scan_namespaces(r, c);
 	nvme_ctrl_scan_paths(r, c);
+	nvme_ctrl_scan_namespaces(r, c);
 	return c;
 }
 
@@ -2262,7 +2309,7 @@ void nvme_rescan_ctrl(struct nvme_ctrl *c)
 		return;
 	nvme_ctrl_scan_namespaces(r, c);
 	nvme_ctrl_scan_paths(r, c);
-	nvme_subsystem_scan_namespaces(r, c->s, NULL, NULL);
+	nvme_subsystem_scan_namespaces(r, c->s);
 }
 
 static int nvme_bytes_to_lba(nvme_ns_t n, off_t offset, size_t count,
@@ -2323,6 +2370,11 @@ const char *nvme_ns_get_sysfs_dir(nvme_ns_t n)
 	return n->sysfs_dir;
 }
 
+const char *nvme_ns_head_get_sysfs_dir(nvme_ns_head_t head)
+{
+	return head->sysfs_dir;
+}
+
 const char *nvme_ns_get_name(nvme_ns_t n)
 {
 	return n->name;
@@ -2749,7 +2801,11 @@ static void nvme_ns_set_generic_name(struct nvme_ns *n, const char *name)
 
 static nvme_ns_t nvme_ns_open(const char *sys_path, const char *name)
 {
+	int ret;
 	struct nvme_ns *n;
+	struct nvme_ns_head *head;
+	struct stat arg;
+	_cleanup_free_ char *path = NULL;
 
 	n = calloc(1, sizeof(*n));
 	if (!n) {
@@ -2757,6 +2813,32 @@ static nvme_ns_t nvme_ns_open(const char *sys_path, const char *name)
 		return NULL;
 	}
 
+	head = calloc(1, sizeof(*head));
+	if (!head) {
+		errno = ENOMEM;
+		free(n);
+		return NULL;
+	}
+
+	head->n = n;
+	list_head_init(&head->paths);
+	ret = asprintf(&path, "%s/%s", sys_path, "multipath");
+	if (ret < 0) {
+		errno = ENOMEM;
+		goto free_ns_head;
+	}
+	/*
+	 * The sysfs-dir "multipath" is available only when nvme multipath
+	 * is configured and we're running kernel version >= 6.14.
+	 */
+	ret = stat(path, &arg);
+	if (ret == 0) {
+		head->sysfs_dir = path;
+		path = NULL;
+	} else
+		head->sysfs_dir = NULL;
+
+	n->head = head;
 	n->fd = -1;
 	n->name = strdup(name);
 
@@ -2765,15 +2847,17 @@ static nvme_ns_t nvme_ns_open(const char *sys_path, const char *name)
 	if (nvme_ns_init(sys_path, n) != 0)
 		goto free_ns;
 
-	list_head_init(&n->paths);
 	list_node_init(&n->entry);
 
 	nvme_ns_release_fd(n); /* Do not leak fds */
+
 	return n;
 
 free_ns:
 	free(n->generic_name);
 	free(n->name);
+free_ns_head:
+	free(head);
 	free(n);
 	return NULL;
 }
@@ -2836,6 +2920,71 @@ nvme_ns_t nvme_scan_namespace(const char *name)
 	return __nvme_scan_namespace(nvme_ns_sysfs_dir(), name);
 }
 
+
+static void nvme_ns_head_scan_path(nvme_subsystem_t s, nvme_ns_t n, char *name)
+{
+	nvme_ctrl_t c;
+	nvme_path_t p;
+
+	nvme_subsystem_for_each_ctrl(s, c) {
+		nvme_ctrl_for_each_path(c, p) {
+			if (!strcmp(nvme_path_get_name(p), name)) {
+				list_add_tail(&n->head->paths, &p->nentry);
+				p->n = n;
+				return;
+			}
+		}
+	}
+}
+
+static void nvme_subsystem_set_ns_path(nvme_subsystem_t s, nvme_ns_t n)
+{
+	struct nvme_ns_head *head = n->head;
+
+	if (nvme_ns_head_get_sysfs_dir(head)) {
+		struct dirents paths = {};
+		int i;
+
+		/*
+		 * When multipath is configured on kernel version >= 6.15,
+		 * we use multipath sysfs link to get each path of a namespace.
+		 */
+		paths.num = nvme_scan_ns_head_paths(head, &paths.ents);
+
+		for (i = 0; i < paths.num; i++)
+			nvme_ns_head_scan_path(s, n, paths.ents[i]->d_name);
+	} else {
+		nvme_ctrl_t c;
+		nvme_path_t p;
+		int ns_ctrl, ns_nsid, ret;
+
+		/*
+		 * If multipath is not configured or we're running on kernel
+		 * version < 6.15, fallback to the old way.
+		 */
+		ret = sscanf(nvme_ns_get_name(n), "nvme%dn%d",
+				&ns_ctrl, &ns_nsid);
+		if (ret != 2)
+			return;
+
+		nvme_subsystem_for_each_ctrl(s, c) {
+			nvme_ctrl_for_each_path(c, p) {
+				int p_subsys, p_ctrl, p_nsid;
+
+				ret = sscanf(nvme_path_get_name(p),
+					     "nvme%dc%dn%d",
+					     &p_subsys, &p_ctrl, &p_nsid);
+				if (ret != 3)
+					continue;
+				if (ns_ctrl == p_subsys && ns_nsid == p_nsid) {
+					list_add_tail(&head->paths, &p->nentry);
+					p->n = n;
+				}
+			}
+		}
+	}
+}
+
 static int nvme_ctrl_scan_namespace(nvme_root_t r, struct nvme_ctrl *c,
 				    char *name)
 {
@@ -2861,37 +3010,13 @@ static int nvme_ctrl_scan_namespace(nvme_root_t r, struct nvme_ctrl *c,
 	n->s = c->s;
 	n->c = c;
 	list_add_tail(&c->namespaces, &n->entry);
+	nvme_subsystem_set_ns_path(c->s, n);
+
 	return 0;
 }
 
-static void nvme_subsystem_set_ns_path(nvme_subsystem_t s, nvme_ns_t n)
-{
-	nvme_ctrl_t c;
-	nvme_path_t p;
-	int ns_ctrl, ns_nsid, ret;
-
-	ret = sscanf(nvme_ns_get_name(n), "nvme%dn%d", &ns_ctrl, &ns_nsid);
-	if (ret != 2)
-		return;
-
-	nvme_subsystem_for_each_ctrl(s, c) {
-		nvme_ctrl_for_each_path(c, p) {
-			int p_subsys, p_ctrl, p_nsid;
-
-			ret = sscanf(nvme_path_get_name(p), "nvme%dc%dn%d",
-				     &p_subsys, &p_ctrl, &p_nsid);
-			if (ret != 3)
-				continue;
-			if (ns_ctrl == p_subsys && ns_nsid == p_nsid) {
-				list_add_tail(&n->paths, &p->nentry);
-				p->n = n;
-			}
-		}
-	}
-}
-
 static int nvme_subsystem_scan_namespace(nvme_root_t r, nvme_subsystem_t s,
-		char *name, nvme_scan_filter_t f, void *f_args)
+		char *name)
 {
 	struct nvme_ns *n, *_n, *__n;
 
@@ -2902,11 +3027,6 @@ static int nvme_subsystem_scan_namespace(nvme_root_t r, nvme_subsystem_t s,
 		nvme_msg(r, LOG_DEBUG, "failed to scan namespace %s\n", name);
 		return -1;
 	}
-	if (f && !f(NULL, NULL, n, f_args)) {
-		nvme_msg(r, LOG_DEBUG, "filter out namespace %s\n", name);
-		__nvme_free_ns(n);
-		return 0;
-	}
 	nvme_subsystem_for_each_ns_safe(s, _n, __n) {
 		struct nvme_path *p, *_p;
 
@@ -2917,7 +3037,7 @@ static int nvme_subsystem_scan_namespace(nvme_root_t r, nvme_subsystem_t s,
 			list_del_init(&p->nentry);
 			p->n = NULL;
 		}
-		list_head_init(&_n->paths);
+		list_head_init(&_n->head->paths);
 		__nvme_free_ns(_n);
 	}
 	n->s = s;
