@@ -263,13 +263,16 @@ static inline void nvme_feature_decode_lba_range(__u32 value, __u8 *num)
 #define NVME_FEAT_TT_TMPTH(v)		NVME_GET(v, FEAT_TT_TMPTH)
 #define NVME_FEAT_TT_TMPSEL(v)		NVME_GET(v, FEAT_TT_TMPSEL)
 #define NVME_FEAT_TT_THSEL(v)		NVME_GET(v, FEAT_TT_THSEL)
+#define NVME_FEAT_TT_TMPTHH(v)		NVME_GET(v, FEAT_TT_TMPTHH)
 
 static inline void nvme_feature_decode_temp_threshold(__u32 value, __u16 *tmpth,
-						      __u8 *tmpsel, __u8 *thsel)
+						      __u8 *tmpsel, __u8 *thsel,
+						      __u8 *tmpthh)
 {
 	*tmpth	= NVME_FEAT_TT_TMPTH(value);
 	*tmpsel	= NVME_FEAT_TT_TMPSEL(value);
 	*thsel	= NVME_FEAT_TT_THSEL(value);
+	*tmpthh	= NVME_FEAT_TT_TMPTHH(value);
 }
 
 #define NVME_FEAT_ER_TLER(v)		NVME_GET(v, FEAT_ERROR_RECOVERY_TLER)
@@ -502,6 +505,9 @@ static inline void nvme_feature_decode_namespace_write_protect(__u32 value,
 	*wps	= NVME_FEAT_WP_WPS(value);
 }
 
+#define NVME_FEAT_BPWPC_BP0WPS(v)	NVME_GET(v, FEAT_BPWPC_BP0WPS)
+#define NVME_FEAT_BPWPC_BP1WPS(v)	NVME_GET(v, FEAT_BPWPC_BP1WPS)
+
 static inline void nvme_id_ns_flbas_to_lbaf_inuse(__u8 flbas, __u8 *lbaf_inuse)
 {
 	*lbaf_inuse = ((NVME_FLBAS_HIGHER(flbas) << 4) |
